#!/usr/bin/scsh -s
!#

;; autogen.scm
;; Copyright (C) 2002, 2003  Julian v. Bock
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License; or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful;
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not; write to the Free Software
;; Foundation; Inc.; 59 Temple Place; Suite 330; Boston; MA  02111-1307  USA
;;
;;    (See file COPYING for more information)


(define (memp pred? v l)
  (let next ((l l))
	(cond ((null? l) #f)
		  ((pred? v (car l)) l)
		  (else (next (cdr l))))))

(define-syntax casep
  (syntax-rules (else)
	((casep pred? (key ...)
			clauses ...)
	 (let ((atom-key (key ...)))
	   (casep pred? atom-key clauses ...)))
	((casep pred? key
			(else result1 result2 ...))
	 (begin result1 result2 ...))
	((casep pred? key
			((atoms ...)
			 result1 result2 ...))
	 (if (memp pred? key '(atoms ...))
		 (begin result1 result2 ...)))
	((casep pred? key
			((atoms ...) result1 result2 ...)
			clause clauses ...)
	 (if (memp pred? key '(atoms ...))
		 (begin result1 result2 ...)
		 (casep pred? key clause clauses ...)))))


(define (usage)
  (for-each (lambda (s)
			  (display s)
			  (newline))
			'("autogen.scm is a script that manages the automatic generation"
			  "of files that are not included in the CVS tree"
			  ""
			  "usage: autogen [option [option...]]"
			  ""
			  "options:"
			  "help       this message"
			  "ref        build api references in the doc directory"
			  "news       create NEWS and TODO"
			  "conf       invoke automake, autoconf etc. and create symlinks"
			  "           in the test directory"
			  "dist       build source distribution"
			  "distcheck  build and check source distribution"
			  "build      build with default options"
			  "pconf      print configure options"
			  "debug      enable debug options"
			  "nodebug    disable debug options"
			  "devel      same as 'autogen debug'"
			  "develb     same as 'autogen devel build'"
			  "develf     same as 'autogen news ref debug'"
			  "develbf    same as 'autogen develf build'"
			  "clean      clean up"
			  "xclean     make sure 'configure' exists before cleaning up"
			  "")))


(if (null? (cdr (command-line)))
	(usage)
	(let ((conf '())
		  (proc-dl #t))
	  (let next ((args (cdr (command-line))))
		(define (run-configure)
		  (if (file-not-exists? "Makefile")
			  (run (./configure ,@conf))))

		(define (run-aclocal)
		  (run (aclocal -I m4)))

		(define (make-ref)
		  (run (doxygen doxygen.config)))

		(define (html2text in out)
		  (run (| (html2text -nobs)
				  (sed s/back_to_main_page//))
			   (< ,in)
			   (> ,out)))

		(define (make-news)
		  (html2text "doc/relnotes.html" "NEWS")
		  (html2text "doc/todo.html" "TODO"))

		(define (make-config)
		  (run (rm -f configure Makefile))
		  (run-aclocal)
		  (run (autoheader))
		  (run (automake -a))
		  (run (autoconf))
		  (run (ln -sf /usr/share/libtool/ltmain.sh)))

		(define (make-config-maybe)
		  (if (file-not-exists? "configure")
			  (make-config)))

		(define (make-debug)
		  (make-config-maybe)
		  (set! conf
				(append '("--enable-debug" "--enable-static" "--disable-shared"
						  "--enable-std-backtraces" "--disable-optimization")
						conf)))

		(define (make-nodebug)
		  (make-config-maybe)
		  (set! conf
				(cons "--disable-debug" conf)))

		(define (make-dist target)
		  (make-clean)
		  (make-news)
		  (make-config)
		  (run-configure)
		  (run (make ,target))
		  (make-clean)
		  (make-ref)
		  (let ((tname (regexp-substitute/global #f
												 (rx "uds-"
													 (submatch (* any))
													 ".tar.")
												 (car (glob "uds-*.tar.gz"))
												 "uds_ref-" 1 ".tar")))
			(run (tar -cvf ,tname -C doc uds_ref))
			(run (bzip2 -kv ,tname))
			(run (gzip -v ,tname)))
		  (make-xclean))

		(define (make-build)
		  (make-config-maybe)
		  (run-configure)
		  (run (make)))

		(define (print-config)
		  (make-config-maybe)
		  (display "configure")
		  (for-each (lambda (x)
					  (display " '")
					  (display x)
					  (display "'"))
					conf)
		  (newline))

		(define (make-clean)
		  (if (file-exists? "configure")
			  (begin (if (file-not-exists? "Makefile")
						 (make-devel))
					 (run (make maintainer-clean)))))

		(define (make-xclean)
		  (make-devel)
		  (make-clean))

		(define (make-devel)
		  (make-config)
		  (make-debug)
		  (run-configure)
		  (run (make TAGS)))

		(define (make-develb)
		  (make-devel)
		  (make-build))

		(define (make-develf)
		  (make-news)
		  (make-ref)
		  (make-devel))

		(define (make-develbf)
		  (make-develf)
		  (make-build))

		(if (not (null? args))
			(begin (casep string=? (car args)
						  (("help" "-h" "--help") (usage))
						  (("ref") (make-ref))
						  (("news") (make-news))
						  (("conf") (make-config))
						  (("debug") (make-debug))
						  (("nodebug") (make-nodebug))
						  (("dist") (make-dist "dist"))
						  (("distcheck") (make-dist "distcheck"))
						  (("build") (make-build))
						  (("pconf") (print-config))
						  (("clean") (make-clean))
						  (("xclean") (make-xclean))
						  (("devel") (make-devel))
						  (("develb") (make-develb))
						  (("develf") (make-develf))
						  (("develbf") (make-develbf))
						  (else (usage)
								(next '())))
				   (next (cdr args)))))))
