/*
 * The UDS Collection
 * Copyright (C) 1999, 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <sys/wait.h>
#include <uds/btrace.hh>
#include <uds/exception.hh>
#include <uds/time.hh>
#include <uds/uds.hh>
#include <csignal>
#include <cstdio>
#include <iostream>
#include <sstream>
#include <unistd.h>

#ifdef _REENTRANT
#include <uds/thread.hh>
#endif


namespace uds
{

namespace
{

void
app_str( stringstream& s, const string& x )
{
	if ( !x.empty() )
	{
		s<<x;
		if ( x[x.size() - 1] != '\n' )
			s<<'\n';
	}
}

}


// Exception implementation

Exception::Exception( bool f, const string& d, const DiagBase& g )
	: addInfo( d ), diag( *g.Clone() )
{
	init();
	if ( f )
		dump();
}

Exception::Exception( const string& d, const DiagBase& g )
	: addInfo( d ), diag( *g.Clone() )
{
	init();
	if ( flags & force_core_dumps )
		dump();
}

Exception::Exception( const Exception& rhs ) : addInfo( rhs.addInfo ),
	btrace( rhs.btrace ), diag( *rhs.diag.Clone() )
{
}

Exception::~Exception()
{
	delete &diag;
}

string
Exception::SourceLocMsg( const char* file, int line, const char* fn )
{
	stringstream str;
	str<<"source: "<<file<<':'<<line<<": "<<fn<<endl;
	return str.str();
}

string
Exception::Diagnose() const
{
	return diag.Diagnose();
}

string
Exception::Message() const
{
	stringstream s;

	if ( !What().empty() )
		s<<What()<<":\n";
	app_str( s, Info() );
	
	string d = Diagnose();
	if ( !d.empty() )
		app_str( s, "diagnose:\n" + d );
	
	s<<btrace;
	
	return s.str();
}

void
Exception::AppMsg( const string& msg )
{
	addInfo += msg;
	if ( addInfo[addInfo.size() - 1] != '\n' )
		addInfo += "\n";
}

Exception&
Exception::operator = ( const Exception& rhs )
{	
	delete &diag;
	
	addInfo = rhs.addInfo;
	diag = *rhs.diag.Clone();

	return *this;
}

void
Exception::init()
{
	if ( flags & std_backtraces )
		btrace = ::uds::Backtrace();
}

void
Exception::dump()
{
	btrace = uds::Backtrace();
#ifdef _REENTRANT
	if ( Thread::Threaded() )
		return;
#endif
	
	fprintf( stderr, "(%i) coredumping...\n", getpid() );
	
	stringstream str;
	str<<"uds_core-"<<getpid()<<'-'<<Time_int64();
	
	int	pid = fork();
	if ( !pid )
	{
		signal( SIGABRT, SIG_DFL );
		abort();
	}
	
	waitpid( pid, 0, 0 );
	
	if ( flags & rename_cores )
		rename( "core", str.str().c_str() );
}

}
