/*
 * The UDS Collection
 * Copyright (C) 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/filestream.hh>
#include <uds/stdexcpt.hh>
#include <fcntl.h>
#include <unistd.h>


namespace uds
{

using namespace std;


namespace
{

int
whence( ios::seekdir way )
{
	switch ( way )
	{
	case ios::beg:
		return SEEK_SET;

	case ios::cur:
		return SEEK_CUR;

	case ios::end:
		return SEEK_END;

	default:
		throw invalid_arg( "bad ios::seekdir" );
	}
}

}


// filebuf implementation

filebuf::filebuf() :
	foff( 0 ), ioff( 0 ), ooff( 0 )
{
}

filebuf::~filebuf()
{
}

filebuf::pos_type 
filebuf::seekoff( off_type off, ios::seekdir dir, ios::openmode mode )
{
	pos_type ret = -1;

	if ( is_open() )
	{
		bool	test_in = mode & ios::in;
		bool	test_out = mode & ios::out;

		if ( test_in )
		{
			switch ( static_cast< int >( dir ) )
			{
			case ios::cur:
				if ( off == 0 )
					return rioff();

				ioff += off;
				break;

			case ios::beg:
				ioff = off;
				break;

			case ios::end:
				ioff = fseek( off, ios::end );
				break;

			default:
				throw invalid_arg( "bad stream direction" );
			}

			setg( libuf(), libuf(), libuf() );
			ret = ioff;
		}
		else if ( test_out )
		{
			if ( dir == ios::cur && off == 0 )
				return rooff();
			
			if ( pbase() < pptr() )
				overflow();
			
			switch ( static_cast< int >( dir ) )
			{
			case ios::cur:
				ooff += off;
				break;

			case ios::beg:
				ooff = off;
				break;

			case ios::end:
				ooff = fseek( off, ios::end );
				break;

			default:
				throw invalid_arg( "bad stream direction" );
			}

			ret = ooff;
		}
	}

	return ret;
}

filebuf::pos_type 
filebuf::seekpos( pos_type pos, ios::openmode mode )
{
	return seekoff( pos, ios::beg, mode );
}

filebuf::int_type
filebuf::Read( char_type* buf, int_type len )
{
	// check whether the output buffer has to be flushed
	
	if ( ( pptr() - pbase() )
		 && !( ( ioff + len < ooff )
			   || ( rooff() < ioff ) ))
		overflow();
	
	if ( foff != ioff )
		fseek( ioff, ios::beg );
	
	int_type ret = fdbuf::Read( buf, len );
	if ( ret >= 0 )
		ioff += ret;

	return ret;
}

filebuf::int_type
filebuf::Write( char_type* buf, int_type len )
{
	if ( foff != ooff )
		fseek( ooff, ios::beg );

	int_type ret = fdbuf::Write( buf, len );
	if ( ret >= 0 )
		ooff += ret;

	return ret;
}

off_t
filebuf::fseek( off_t off, ios::seekdir dir )
{
	foff = -1;

	if ( is_open() )
		foff = lseek( File(), off, whence( dir ) );

	return foff;
}


// FileStream implementation

FileStream::FileStream() :
	iofdstream( &buf ), pfile( false )
{
}

FileStream::FileStream( int fd, ios::openmode mode, bool pf ) :
	iofdstream( &buf )
{
	Open( fd, mode, pf );
}

FileStream::FileStream( const char* fn, ios::openmode mode,
						efs_flags_t ef, mode_t prot, bool pf ) :
	iofdstream( &buf )
{
	Open( fn, mode, ef, prot, pf );
}

FileStream::FileStream( const string& fn, ios::openmode mode,
						efs_flags_t ef, mode_t prot, bool pf ) :
	iofdstream( &buf )
{
	Open( fn, mode, ef, prot, pf );
}

FileStream::~FileStream()
{
	if ( !pfile )
		Close();
}

void
FileStream::Open( int fd, ios::openmode mode, bool pf )
{
	if ( !pfile )
		Close();
	pfile = pf;

	om_sanity_check( mode );
	buf.Attach( fd, mode );
}

void
FileStream::Open( const char* fn, ios::openmode mode,
				  efs_flags_t ef, mode_t prot, bool pf )
{
	if ( !pfile )
		Close();
	pfile = pf;

	om_sanity_check( mode );

	int flags = ( mode & ios::out ) ? O_RDWR : O_RDONLY;

	if ( mode & ios::app )
		flags |= O_APPEND;
	if ( mode & ios::trunc )
		flags |= ( O_TRUNC | O_CREAT );
	if ( ef & excl )
		flags |= O_EXCL;
	if ( ef & create )
		flags |= O_CREAT;

	int fd = open( fn, flags, prot );
	if ( fd < 0 )
		throw system_error( string( "error while opening file " ) + fn );
	buf.Attach( fd, mode );

	if ( mode & ios::ate )
		buf.seekoff( 0, ios::end, mode );
}

void
FileStream::Close()
{
	if ( IsOpen() )
	{
		int fd = File();
		buf.Detach();
		close( fd );
	}
}

void
FileStream::om_sanity_check( ios::openmode mode )
{
	int iom = mode & ( ios::in | ios::out );
	if ( !iom )
		throw file_error( "neither ios::in nor ios::out set" );
}

}
