/*
 * The UDS Collection
 * Copyright (C) 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/io_util.hh>
#include <uds/sys_util.hh>
#include <sstream>


namespace uds
{

// misc. implementation

namespace
{

template < class In >
size_t
bs_read( In& in, string& v, brw_flags_t f, endianess_t e, size_t ml )
{
	const size_t buf_size = 2 << 10;
	char buf[buf_size];

	v.erase();

	size_t size;
	size_t cg = BRead( in, size, f, e );
	if ( cg != sizeof( size_t ) )
		throw brw_string_len_error( 4, 4, cg );

	size_t rs = ( size > ml ) ? ( ( f & brw_readps ) ? ml : 0 ) : size;

	while ( rs )
	{
		size_t nreq = rs <? buf_size;
		cg = BReadRaw( in, buf, nreq, f );
		rs -= cg;

		if ( cg != nreq )
			throw brw_string_len_error( ml, size, v.size() + cg );

		v.append( buf, cg );
	}

	if ( size > v.size() )
	{
		if ( f & brw_skip )
			for ( size_t ic = size - v.size(); ic; )
			{
				size_t nreq = ic <? buf_size;
				cg = BReadRaw( in, buf, nreq, f );
				ic -= cg;

				if ( cg != nreq )
					throw brw_string_len_error( ml, size, size - ic );
			}

		if ( ~f & brw_lnothrow )
			throw brw_string_len_error( ml, size,
										( f & brw_skip ) ? size : v.size() );
	}

	return v.size();
}

template < class Out >
inline void
bs_write( Out& out, const string& v, endianess_t e )
{
	BWrite( out, v.size(), e );
	BWriteRaw( out, v.data(), v.size() );
}

}


// brw_string_len_error implementation

brw_string_len_error::brw_string_len_error( size_t ml, size_t rl, size_t br ) :
	runtime_error( string(), uds::NoDiag() ), maxlen( ml ), rlen( rl ),
	bread( br )
{
	stringstream ss;
	ss<<"max. length: "<<ml<<" requested length: "<<rl<<" bytes read: "<<br;

	AppMsg( ss.str() );
}

brw_string_len_error::~brw_string_len_error()
{
}


// misc. implementation

void
ReadLine( istream& in, string& o, bool app, bool ign, char eol )
{
	static const size_t buf_size = 2 << 9;
	char buf[buf_size];
	
	o.erase();
	if ( !in.good() )
		return;
	
	do
	{
		in.get( buf, buf_size, eol );
		o.append( buf, in.gcount() );
	} while ( in.gcount() == buf_size - 1 && in.peek() != eol );

	if ( o.empty() )
		in.clear( in.rdstate() & ~ios::failbit );
	
	if ( app )
		o.append( 1u, in.peek() );
	
	if ( ign )
		in.get();
}

string
ReadLine( istream& in, bool app, bool ign, char eol )
{
	string ret;
	ReadLine( in, ret, app, ign, eol );
	return ret;
}

size_t
BReadRaw( ifdstream& in, void* buf, size_t l, brw_flags_t flags )
{
	errno = 0;
	in.read( static_cast< char* >( buf ), l );
	if ( in.bad() && errno )
		throw system_error();
	
	if ( flags & brw_wait && in.gcount() < static_cast< streamsize >( l ) )
	{
		size_t rl = in.gcount();
	
		FD_Set fds;
		fds.Set( in.File() );

		while ( rl < l )
		{
			int e = select( fds.N(), fds(), 0, 0, 0 );
			in.clear();
			
			if ( e < 0 )
			{
				if ( errno == EINTR )
					continue;
				Throw( system_error( "error while waiting for input" ) );
			}
			
			in.read( static_cast< char* >( buf ) + rl, l - rl );
			if ( !in.gcount() )
				Throw( runtime_error( "connection lost" ) );
			
			rl += in.gcount();
		}
		
		return l;
	}
	
	return in.gcount();
}

size_t
BRead( ifdstream& in, string& v, brw_flags_t f, endianess_t end, size_t ml )
{
	return bs_read( in, v, f, end, ml );
}

size_t
BReadRaw( int in, void* buf, size_t l, brw_flags_t f )
{
	size_t gc = CkStdCall( read( in, static_cast< char* >( buf ), l ) );
	
	if ( f & brw_wait && gc < l )
	{
		size_t rl = gc;
	
		FD_Set fds;
		fds.Set( in );
		
		while ( rl < l )
		{
			int e = select( fds.N(), fds(), 0, 0, 0 );
			
			if ( e < 0 )
			{
				if ( errno == EINTR )
					continue;
				Throw( system_error( "error while waiting for input" ) );
			}
			
			gc = CkStdCall( read( in, static_cast< char* >( buf ) + rl,
								  l - rl ) );
			if ( !gc )
				Throw( runtime_error( "connection lost" ) );
			
			rl += gc;
		}
		
		return l;
	}
	
	return gc;
}

size_t
BRead( int in, string& v, brw_flags_t f, endianess_t e, size_t ml )
{
	return bs_read( in, v, f, e, ml );
}

void
BWrite( ostream& out, const string& v, endianess_t end )
{
	bs_write( out, v, end );
}

void
BWrite( int out, const string& v, endianess_t end )
{
	bs_write( out, v, end );
}

}
