/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <uds/procstream.hh>
#include <uds/stdexcpt.hh>
#include <uds/sys_util.hh>
#include <uds/thread.hh>
#include <algorithm>
#include <procstream.hh>
#include <set>
#include <unistd.h>


#ifndef PF_LOCAL
#define PF_LOCAL PF_UNIX
#endif


namespace uds
{

// ProcStream implementation

namespace
{

typedef set< int > proc_fdl_t;

proc_fdl_t&
proc_fdl()
{
	static char mem[sizeof( proc_fdl_t )];
	static proc_fdl_t* o = new ( mem ) proc_fdl_t();
	return *o;
}

Mutex&
proc_fdl_mutex()
{
	static char mem[sizeof( Mutex )];
	static Mutex* o = new ( mem ) Mutex();
	return *o;
}

}

ProcStream::ProcStream() :
	iostream( &sbuf ), sock( -1 ), cpid( 0 )
{
}

ProcStream::ProcStream( const char* cmd, bool t, int s, const char** env )
	: iostream( &sbuf ), sock( -1 ), cpid( 0 )
{
	Open( cmd, t, s, env );
}

ProcStream::ProcStream( const string& cmd, bool t, int s, const char** env )
	: iostream( &sbuf ), sock( -1 ), cpid( 0 )
{
	Open( cmd.c_str(), t, s, env );
}

ProcStream::~ProcStream()
{
	Close();
}

void
ProcStream::Open( const char* cmd, bool t, int s, const char** env )
{
	Close();
	term = t;
	sig = s;
	
	int socks[2];
	CkStdCall( socketpair( PF_LOCAL, SOCK_STREAM, 0, socks ) );
	
	if ( ( cpid = Fork() ) == 0 )
	{
		::close( socks[0] );
		int sd = socks[1];
		
		if ( sd != 0 )
			CkStdCall( dup2( sd, 0 ) );
		if ( sd != 1 )
			CkStdCall( dup2( sd, 1 ) );
		
		::close( sd );
		
		// close streams from previous ProcStream::Open calls
		MutexLock lock( proc_fdl_mutex() );
		for_each( proc_fdl().begin(), proc_fdl().end(), &::close );
		
		// execute program
		if ( env )
			execle( "/bin/sh", "sh", "-c", cmd, 0, env );
		else
			execl( "/bin/sh", "sh", "-c", cmd, 0 );
		
		throw uds::system_error( string( "unable to execute command: "
										 "/bin/sh -c " ) += cmd );
	}
	
	::close( socks[1] );
	sock = socks[0];
	
	{
		MutexLock lock( proc_fdl_mutex() );
		proc_fdl().insert( sock );
	}
	
	sbuf.Attach( sock );
}

void
ProcStream::Close()
{
	if ( sock < 0 )
		return;

	sbuf.Detach();

	if ( term && waitpid( cpid, 0, WNOHANG ) == 0 )
		kill( cpid, sig );
	waitpid( cpid, 0, 0 );
	cpid = 0;
	
	MutexLock lock( proc_fdl_mutex() );
	
	proc_fdl().erase( sock );
	sock = -1;	
}

}
