/*
 * The UDS Collection
 * Copyright (C) 1999, 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/alloc.hh>
#include <uds/btrace.hh>
#include <uds/init.hh>
#include <uds/idep.hh>
#include <uds/io_util.hh>
#include <uds/procstream.hh>
#include <uds/sentinel.hh>
#include <uds/stdexcpt.hh>
#include <uds/str_util.hh>
#include <uds/uds.hh>
#include <cstdio>
#include <iomanip>
#include <sstream>


namespace uds
{

void
throw_zombie( void* fool )
{
	stringstream msg;
	msg<<"virtual function call failed (fool: "<<fool<<")\n";
	
	if ( uncaught_exception() )
		cerr<<msg.str()<<Backtrace();
	else
		Throw( zombie_object( msg.str() ));
}

const int ret_depth = 1;
const int zt_size = 512;

typedef void ( *z_type )( void* );
z_type zombie_table[zt_size];


// Sentinel implementation

namespace
{

inline bool
log()
{
	return flags & log_allocs;
}

}


void*
Sentinel::Allocate( size_t size )
{
	if ( Initialized() )
	{
		MutexLock lock( mut );

		if ( log() )
		{
			char buf[100];
			int c = sprintf( buf, "function at address %p allocates %u "
							 "bytes at address... ",
							 __builtin_return_address( ret_depth ), size );
			write( 2, buf, c );
		}
	
		void* p = original_new( size );
		if ( !p )
			throw bad_alloc();

		heap_map.insert(
			make_pair( p, mem_t( __builtin_return_address( ret_depth ),
								 size ) ));

		if ( log() )
		{
			char buf[20];
			int c = sprintf( buf, "%p\n", p );
			write( 2, buf, c );
		}

		return p;
	}

	return original_new( size );
}

void
Sentinel::Release( void* p )
{
	if ( Initialized() )
	{
		MutexLock lock( mut );
	
		if ( log() )
		{
			char buf[100];
			int c = sprintf( buf, "%p releases memory at %p...",
							 __builtin_return_address( ret_depth ), p );
			write( 2, buf, c );
		}
	
		heap_map_t::iterator i = heap_map.find( p );
		if ( i == heap_map.end() )
		{
			if ( !uncaught_exception() )
				throw zombie_object( "operator delete failed - "
									 "memory was not allocated\n" );
			return;
		}
	
		int s = i->second.size;
		if ( flags & summon_zombies && s % sizeof( void* ) == 0 )
		{
			void** q = static_cast< void** >( p );
			void** e = reinterpret_cast< void** >( static_cast< char* >( p ) + s );
		
			while ( q < e )
				*q++ = reinterpret_cast< void* >( zombie_table );
		}
		else
			memset( p, 0, s );

		if ( log() )
		{
			char buf[100];
			int c = sprintf( buf, "%i bytes of memory allocated by %p freed\n",
							 s, i->second.caller );
			write( 2, buf, c );
		}

		heap_map.erase( i );
	}

	free( p );
}

bool
Sentinel::Allocated( const void* p ) const
{
	MutexLock lock( mut );
	
	if ( heap_map.empty() )
		return false;

	heap_map_t::const_iterator i = --heap_map.lower_bound( p );
	return static_cast< const char* >( i->first ) + i->second.size > p;
}

void
Sentinel::Check()
{
	int				nleaks = 0;
	stringstream	ss;
	string			func;
	string			sline;
	string			msg;
	ProcStream		pr;

	Addr2LinePS( pr, false );
	
	for ( heap_map_t::iterator i = heap_map.begin();
		  i != heap_map.end(); ++i )
	{
		if ( pr.Open() )
		{
			pr<<i->second.caller<<endl;
			ReadLine( pr, func );
			ReadLine( pr, sline );

			bl_set_t::const_iterator j = blist.begin();
			for ( ; j != blist.end(); ++j )
			{
				if ( sline.find( *j ) != string::npos )
					break;
			}
			if ( j != blist.end() )
				continue;

			size_t ls = sline.rfind( '/' );
			if ( ls != string::npos )
				sline.erase( 0, ls + 1 );

			if ( sline == "??:0" )
			{
				if ( ignu )
					continue;
				
				sline = StringVar( i->second.caller );
			}
		}
		else
			func = StringVar( i->second.caller );

		++nleaks;

		ss<<left<<setw( 6 )<<i->second.size<<' '
		  <<setw( 10 )<<i->first<<' '
		  <<left<<setw( 18 )<<sline<<' ';
		ss<<BreakLine( func, 37, 79, 79 - ss.str().size() );

		msg += ss.str();
		ss.str( "" );
	}

	if ( eleaks.find( nleaks ) != eleaks.end() )
		return;
	
	ss<<"\nWARNING: "<<nleaks<<" memory leak[s] detected\n"
	  <<setw( 6 )<<"size"<<' '
	  <<setw( 10 )<<"address"<<' '
	  <<setw( 18 )<<"source"<<' '
	  <<"function\n"
	  <<string( 79, '-' )<<endl;

	write( 2, ss.str().data(), ss.str().size() );
	write( 2, msg.data(), msg.size() );
}

Sentinel::Sentinel() :
	ignu( false )
{
	eleaks.insert( 0 );
	
	for ( int i = 0; i < zt_size; ++i )
		zombie_table[i] = throw_zombie;
}

Sentinel::~Sentinel()
{
}

}
