/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include "config.h"
#ifdef HAVE_SOCKET

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <uds/stdexcpt.hh>
#include <uds/streamsocket.hh>
#include <unistd.h>


#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif
#ifndef PF_LOCAL
#define PF_LOCAL PF_UNIX
#endif
#ifndef SUN_LEN
#define SUN_LEN( ptr ) ( ( size_t )(( ( sockaddr_un* )0 )->sun_path )        \
	+ strlen( ( ptr )->sun_path ) )
#endif


namespace uds
{

// sockbuf implementation

sockbuf::sockbuf() :
	sf( 0 )
{
}

sockbuf::~sockbuf()
{
}

sockbuf::int_type
sockbuf::Read( char_type* buf, int_type len )
{
	while ( true )
	{
		int rlen = recv( File(), buf, len, sf );

		if ( rlen < 0 )
		{
			if ( errno == EINTR )
				continue;
			else if ( errno != EAGAIN )
				Throw( system_error( "stream buffer underflow: "
									 "recv() failed" ) );

			return 0;
		}

		return rlen;
	}
}

sockbuf::int_type
sockbuf::Write( char_type* buf, int_type len )
{
	while ( true )
	{
		int rlen = send( File(), buf, len, sf );
			
		if ( rlen < 0 )
		{
			if ( errno == EINTR )
				continue;
			else if ( errno != EAGAIN )
				Throw( system_error( "stream buffer overflow: "
									 "send() failed" ) );

			return 0;
		}
		
		return rlen;
	}
}


// StreamSocket implementation

StreamSocket::StreamSocket()
	: iofdstream( &sb )
{
}

StreamSocket::StreamSocket( int fd )
	 : iofdstream( &sb )
{
	Attach( fd );
}

StreamSocket::~StreamSocket()
{
	priv_close();
}

void
StreamSocket::Close()
{
	priv_close();	
	Socket_base::Close();
}

void
StreamSocket::Attach( int s )
{
	Socket_base::Attach( s );
	sb.Attach( s );
}

void
StreamSocket::priv_close()
{
	if ( Valid() )
		close( Sock() );
}


// TCPSocket implementation

TCPSocket::TCPSocket()
{
}

TCPSocket::TCPSocket( int fd ) : StreamSocket( fd )
{
}

TCPSocket::~TCPSocket()
{
}

void
TCPSocket::Listen( unsigned short p, unsigned n )
{
	Attach( CrTCPSocket( p ) );
	Socket_base::Listen( n );
}

void
TCPSocket::Connect( const char* hn, unsigned short p )
{
	Attach( CkStdCall( socket( PF_INET, SOCK_STREAM, 0 ) ));
		
	sockaddr_in name;
	init_inet_addr( name, hn, p );
	CkStdCall( connect( Sock(), reinterpret_cast< sockaddr* >( &name ),
						sizeof( name ) ));
}


// UnixSocket implementation

UnixSocket::UnixSocket()
{
}

UnixSocket::UnixSocket( int fd ) :
	StreamSocket( fd )
{
}

UnixSocket::~UnixSocket()
{
}

void
UnixSocket::Listen( const char* f, unsigned n )
{
	Attach( CrUnixSocket( f ) );
	Socket_base::Listen( n );
}

void
UnixSocket::Connect( const char* n )
{
	Attach( CkStdCall( socket( PF_LOCAL, SOCK_STREAM, 0 ) ));
		
	sockaddr_un name;
	init_local_addr( name, n );
	CkStdCall( connect( Sock(), reinterpret_cast< sockaddr* >( &name ),
						SUN_LEN( &name ) ));
}


// misc. implementation

StreamSocket&
sockflags_fn( StreamSocket& o, int f )
{
	o.SFlags( f );
	return o;
}

int
CrTCPSocket( unsigned short p )
{
	// create the socket
	int sock = CkStdCall( socket( PF_INET, SOCK_STREAM, 0 ) );
	
	// bind a name to the socket
	sockaddr_in	name;
	init_inet_addr( name, p );
	CkStdCall( bind( sock, reinterpret_cast< sockaddr* >( &name ),
					 sizeof( name ) ));
	
	return sock;	
}

int
CrUnixSocket( const char *filename )
{
	// create the socket
	int sock = CkStdCall( socket( PF_LOCAL, SOCK_STREAM, 0 ) );
	
	// bind a name to the socket
	sockaddr_un name;
	init_local_addr( name, filename );
	CkStdCall( bind( sock, reinterpret_cast< sockaddr* >( &name ),
					 SUN_LEN( &name ) ));
	
	return sock;
}

}

#endif
