/*
 * The UDS Collection
 * Copyright (C) 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include "config.h"
#if defined HAVE_LINUX_WANPIPE_H && defined HAVE_LINUX_IF_WANPIPE_H \
	&& defined HAVE_LINUX_SDLA_X25_H

#include <linux/if_ether.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <uds/sys_util.hh>
#include <uds/swanpipe.hh>
#include <cerrno>
#include <string>
#include <sstream>


namespace uds
{

namespace
{

void
init_wanpipe_addr( wan_sockaddr_ll& sa, const char* ifc, bool svr )
{
	sa.sll_family = AF_WANPIPE;
	sa.sll_protocol = htons( X25_PROT );
	strcpy( reinterpret_cast< char * >( sa.sll_device ),
			svr ? "svc_listen" : "svc_connect" );
	strcpy( reinterpret_cast< char * >( sa.sll_card ), ifc );
}

}


// X25Api_t implementation

X25Api_t::X25Api_t( bool c )
{
	if ( c )
		Clear();
}

X25Api_t::X25Api_t( const string& d, unsigned char qdm )
{
	Clear();
	SetData( d, qdm );
}

X25Api_t::~X25Api_t()
{
}

void
X25Api_t::SetData( const string& d, unsigned char qdm )
{
	if ( d.size() > sizeof( api.data ) )
		 throw runtime_error( "data too long for x25api_t structure" );
		 
	api.hdr.length = d.size();
	memcpy( api.data, d.data(), d.size() );

	api.hdr.qdm = qdm;
}

void
X25Api_t::Send( int sock, bool oob ) const
{
	send( sock, &api, api.hdr.length + sizeof( x25api_hdr_t ),
		  oob ? MSG_OOB : 0 );
}


ostream&
operator << ( ostream& out, const X25Api_t& x )
{
	out<<"type: "<<static_cast< int >( x->hdr.pktType )
	   <<"\nqdm: "<<static_cast< int >( x->hdr.qdm )
	   <<"\ncause: "<<static_cast< int >( x->hdr.cause )
	   <<"\ndiagnose: "<<static_cast< int >( x->hdr.diagn )
	   <<"\nresult: "<<static_cast< int >( x->hdr.result )
	   <<"\nlength: "<<static_cast< int >( x->hdr.length )
	   <<"\nLCN: "<<static_cast< int >( x->hdr.lcn )<<endl;

	return out;
}


// X25_SWanpipe implementation

X25_SWanpipe::X25_SWanpipe()
{
}
  
X25_SWanpipe::X25_SWanpipe( int fd ) :
	PacketSocket( fd )
{
}
	
X25_SWanpipe::~X25_SWanpipe()
{
	priv_close();
}

void
X25_SWanpipe::Close()
{
	priv_close();
	Socket_base::Close();
}

void
X25_SWanpipe::Listen( const char* ifc, unsigned n )
{
	Attach( CrX25_SWanpipeSocket( ifc ) );
	Socket_base::Listen( n );
}

int
X25_SWanpipe::Accept()
{
	int sock = PacketSocket::Accept();
	CkStdCall( ioctl( Sock(), SIOC_WANPIPE_ACCEPT_CALL, 0 ) );

	return sock;
}

void
X25_SWanpipe::Connect( const char* ifc, size_t ps, const char* sn,
					   const char* dn )
{
	Attach( CrX25_SWanpipeSocket( ifc, false ) );

	stringstream cmd;
	cmd<<"-d"<<dn<<" -s"<<sn;

	X25Api_t api_cmd( cmd.str() );
	CkStdCall( ioctl( Sock(), SIOC_WANPIPE_SET_CALL_DATA, &api_cmd ) );

	wan_sockaddr_ll sa;
	init_wanpipe_addr( sa, ifc, false );
	CkStdCall( connect( Sock(), reinterpret_cast< sockaddr* >( &sa ),
						sizeof( sa ) ));

	CkStdCall( ioctl( Sock(), SIOC_WANPIPE_GET_CALL_DATA, &api_cmd ) );
	lcn = api_cmd->hdr.lcn;
	pack_sz = ps;
}

void
X25_SWanpipe::Recv( string& data )
{
	X25Api_t x;

	do
	{
		FD_Set fds;
		fds.Set( Sock() );

		int ret = select( fds.N(), fds(), 0, 0, 0 );
		if ( ret < 0 )
		{
			if ( errno == EINTR )
				continue;
			throw system_error();
		}

		Recv( x );
		data += x.Data();
	} while ( x->hdr.qdm & X25CMD_M_BIT_MASK );
}

void
X25_SWanpipe::Recv( X25Api_t& p, bool oob )
{
	p.Clear();
	CkStdCall( recv( Sock(), p(), sizeof( x25api_t ), oob ? MSG_OOB : 0 ) );
}

void
X25_SWanpipe::Send( const string& d, bool oob )
{
	X25Api_t x;
	size_t pos = 0;

	for ( ; pos < d.size(); pos += pack_sz )
	{
		size_t delta = d.size() - pos;
		x.SetData( d.substr( pos, delta <? pack_sz ),
				   delta > pack_sz ? X25CMD_M_BIT_MASK : 0 );
		
		x.Send( Sock(), oob );
	}
}

bool
X25_SWanpipe::Link() const
{
    X25Api_t x( true );
	return ioctl( Sock(), SIOC_WANPIPE_SOCK_STATE, x() );
}

void
X25_SWanpipe::priv_close()
{
	if ( Valid() )
	{
		x25api_t x;
		memset( &x, 0, sizeof( x25api_t ) );
		ioctl( Sock(), SIOC_WANPIPE_RESET_CALL, &x );
	}
}


// misc. implementation

int
CrX25_SWanpipeSocket( const char* ifc, bool svr )
{
	int sock = CkStdCall( socket( AF_WANPIPE, SOCK_RAW, 0 ) );

	wan_sockaddr_ll sa;
	init_wanpipe_addr( sa, ifc, svr );
	CkStdCall( bind( sock, reinterpret_cast< sockaddr* >( &sa ),
					 sizeof( wan_sockaddr_ll ) ));

	return sock;
}

}

#endif // HAVE_LINUX_WANPIPE_H
