/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <uds/childdiag.hh>
#include <uds/exit.hh>
#include <uds/filestream.hh>
#include <uds/random.hh>
#include <uds/sys_util.hh>
#include <cerrno>
#include <cstdio>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <sstream>
#include <unistd.h>


namespace uds
{

namespace
{

int
tmp_file( const string& pre, string* tmp )
{
	static const char* chars = "abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	static Random random;
	
	random.Seed( TimeSeed(), true );
	int c = 16;
	
	do
	{
		stringstream str;
		str<<pre<<'-';
		
		unsigned rand = random();
		
		for ( int j = 0; j < 6; ++j )
		{
			str<<chars[rand % 62];
			rand /= 62;
		}
		
		int fd = open( str.str().c_str(), O_RDWR | O_CREAT | O_EXCL, 0600 );
		if ( fd >= 0 )
		{
			if ( tmp )
				*tmp = str.str();
			return fd;
		}
		
		--c;
	} while ( c );
	
	throw runtime_error( "unable to find unique temporary file name" );
}

void
ck_group( group* grp )
{
	if ( !grp )
	{
		if ( errno )
			throw system_error();
             
		throw runtime_error( "group does not exist" );
	}
}

void
ck_user( passwd* usr )
{
	if ( !usr )
	{
		if ( errno )
			throw system_error();
             
		throw runtime_error( "user does not exist" );
	}
}

}


string TmpFileName_Prefix __attribute__(( weak )) = "/tmp/uds";


// FD_Set implementation

void
FD_Set::Set( int fd )
{
	FD_SET( fd, &fds );
	n = n >? ( fd + 1 );
}

void
FD_Set::Clear( int fd )
{
	FD_CLR( fd, &fds );
}

void
FD_Set::Zero()
{
	FD_ZERO( &fds );
	n = 1;
}

bool
FD_Set::IsSet( int fd ) const
{
	return FD_ISSET( fd, &fds );
}


// misc. implementation

bool
FileExist( const char* f )
{
	struct stat	s;
	
	if ( !stat( f, &s ) )
		return true;
	if ( errno == ENOENT )
		return false;
	
	throw file_error( "unable to determine state of file "
					  + string( f ) + '\n' );
}

bool
FileAccess( const char* f, int m )
{
	return access( f, m ) == 0;
}

string
CurrentWD()
{
	char* buf = getcwd( 0, 0 );
	if ( !buf )
		Throw( system_error( "unable to determine current "
							 "working directory" ) );
	
	string ret = buf;
	free( buf );
	return ret;
}

string
TmpFileName( const string& pre )
{
	string name;
	int fd = tmp_file( pre, &name );
	close( fd );
	
	return name;
}

string
TmpFileName()
{
	return TmpFileName( TmpFileName_Prefix );
}

int
TmpFile( const string& pre )
{
	return tmp_file( pre, 0 );
}

int
TmpFile()
{
	return TmpFile( TmpFileName_Prefix );
}

int
Fork()
{
	int pid = fork();
	
	if ( pid == -1 )
		Throw( fork_error() );
	return pid;
}

int
Wait( int pid, int iv, int opt )
{
	int	status;
	int	p = CkStdCall( waitpid( pid, &status, opt ),
					   "call to wait() failed\n" );
	
	if ( p && ( !WIFEXITED( status ) || WEXITSTATUS( status ) ))
	{
		if ( WEXITSTATUS( status ) != iv )
			Throw( child_error( "", ChildDiag( status ) ));	
		else Exit( iv );
	}
	
	return p;
}

int
MkLockFile( const char* lockfile, bool create, pid_t* lpid )
{
	// look for existing lockfile

	if ( lpid )
		*lpid = -1;

	int lfd = open( lockfile, create ? O_RDWR : O_RDONLY );
	if ( lfd >= 0 )
	{
		// check whether another process holds a lock on the file
		
		if ( flock( lfd, LOCK_EX | LOCK_NB ) )
		{
			if ( lpid )
			{
				FileStream lfs( lfd, ios::in | ios::out, true );
				*lpid = 0;
				lfs>>*lpid;
			}

			close( lfd );
			return -1;
		}

		return lfd;
	}
		
	// lockfile does not exist or access is denied

	if ( create )
	{
		lfd = open( lockfile,
					O_RDWR | O_CREAT | O_EXCL,
					S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );

		bool ofail = lfd < 0;

		if ( ofail || flock( lfd, LOCK_EX | LOCK_NB ) < 0 )
		{
			if ( !ofail )
				close( lfd );
			
			return -1;
		}
	}
	
	return lfd;
}

void
WriteLockFile( int lfd )
{
	CkStdCall( lseek( lfd, 0, SEEK_SET ) );
	CkStdCall( ftruncate( lfd, 0 ) );

	stringstream s;
	s<<getpid()<<endl;
	CkStdCall( write( lfd, s.str().data(), s.str().size() ));
}

void
System( const char* cmd )
{
	pid_t cpid = fork();
	if ( cpid == 0 )
	{
		execl( "/bin/sh", "sh", "-c", cmd, 0 );
		
		throw uds::system_error( string( "unable to execute command: "
										 "/bin/sh -c " ) += cmd );
	}

	Wait( cpid );
}

string
GroupName( gid_t gid )
{
	errno = 0;
	group* grp = getgrgid( gid );
	ck_group( grp );
		
	return grp->gr_name;
}

gid_t
GroupID( const char* name )
{
	errno = 0;
	group* grp = getgrnam( name );
	ck_group( grp );
		
	return grp->gr_gid;
}

string
UserName( uid_t uid )
{
	errno = 0;
	passwd* usr = getpwuid( uid );
	ck_user( usr );
		
	return usr->pw_name;
}

uid_t
UserID( const char* name )
{
	errno = 0;
	passwd* usr = getpwnam( name );
	ck_user( usr );
		
	return usr->pw_uid;
}

}
