/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002, 2003  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/exit.hh>
#include <uds/functional.hh>
#include <uds/sentinel.hh>
#include <uds/stl_ext.hh>
#include <uds/stl_util.hh>
#include <uds/thread.hh>
#include <map>


namespace uds
{

// Mutex implementation

void
Mutex::Lock() const
{
	if ( pthread_mutex_lock( &mut ) )
		Throw( system_error( "unable to lock mutex" ) );
}

void
Mutex::UnLock() const
{
	if ( pthread_mutex_unlock( &mut ) )
		Throw( system_error( "unable to unlock mutex" ) );
}


// CMutexLock implementation

CMutexLock::CMutexLock( const Mutex& m ) : mut( m )
{
	Thread::Self()->PushCleanupHandler(
		Thread::CleanupHandler( &pthread_mutex_unlock, &mut.Get() ));
	mut.Lock();
}

CMutexLock::~CMutexLock()
{
	Thread::Self()->PopCleanupHandler( true );
}


// CondVar implementation

CondVar::CondVar()
{
	pthread_cond_init( &cond, 0 );
}

CondVar::~CondVar()
{
	pthread_cond_destroy( &cond );
}


// Semaphore implementation

Semaphore::Semaphore( bool s, int v )
{
	CkStdCall( sem_init( &sem, s, v ), "unable to initialize semaphore" );
}

Semaphore::~Semaphore()
{
	CkStdCall( sem_destroy( &sem ), "unable to destroy semaphore" );
}

void
Semaphore::Post() const
{
	CkStdCall( sem_post( &sem ), "unable to increase semaphore count" );
}

int
Semaphore::Value() const
{
	int v;
	sem_getvalue( &sem, &v );
	return v;
}


// ThreadAttr implementation

ThreadAttr::ThreadAttr()
{
	pthread_attr_init( &attr );
}

ThreadAttr::~ThreadAttr()
{
	pthread_attr_destroy( &attr );
}

int
ThreadAttr::DetachState() const
{
	int v;
	pthread_attr_getdetachstate( &attr, &v );
	return v;
}

int
ThreadAttr::SchedPolicy() const
{
	int v;
	pthread_attr_getschedpolicy( &attr, &v );
	return v;
}

int
ThreadAttr::SchedParam() const
{
	sched_param v;
	pthread_attr_getschedparam( &attr, &v );
	return v.sched_priority;
}

void
ThreadAttr::SchedParam( int v )
{
	sched_param p;
	p.sched_priority = v;
	pthread_attr_setschedparam( &attr, &p );
}

int
ThreadAttr::InheritSched() const
{
	int v;
	pthread_attr_getinheritsched( &attr, &v );
	return v;
}

int
ThreadAttr::Scope() const
{
	int v;
	pthread_attr_getscope( &attr, &v );
	return v;
}


// Thread implementation

typedef map< pthread_t, Thread* > TIMap;

namespace
{

TIMap&
ThreadMap()
{
	static TIMap m;
	return m;
}

Mutex&
ThreadMapMutex()
{
	static Mutex m;
	return m;
}

}

void
Thread::Cancel()
{
	if ( pthread_cancel( thread ) )
		Throw( system_error( "unable to cancel thread" ) );
}

void*
Thread::Join( const Thread& th )
{
	void* ret;
	if ( pthread_join( th.Get(), &ret ) )
		Throw( system_error( "unable to join thread" ) );
	
	return ret;
}

int
Thread::SetCancelState( int t )
{
	int old;
	if ( pthread_setcancelstate( t, &old ) )
		Throw( system_error( "unable to set cancel state" ) );
	return old;
}

int
Thread::SetCancelType( int t )
{
	int old;
	if ( pthread_setcanceltype( t, &old ) )
		Throw( system_error( "unable to set cancel type" ) );
	return old;
}

void
Thread::Detach()
{
	if ( pthread_detach( thread ) )
		Throw( system_error( "unable to detach thread" ) );
}

void
Thread::PushCleanupHandler( const CleanupHandler& ch )
{
	MutexLock lock( mutex );
	chs.push( ch );
}

void
Thread::PopCleanupHandler( bool e )
{
	MutexLock lock( mutex );
	if ( e )
		chs.top()();
	chs.pop();
}

Thread::GC_Ptr
Thread::Self()
{
	MutexLock	lock( ThreadMapMutex() );
	pthread_t	self = pthread_self();
	
	Thread* t = SafeDeref( MapQueryC( ThreadMap(), self ) );
	if ( !t )
		t = new Thread( self );
	
	return t;
}

bool
Thread::Threaded()
{
	MutexLock lock( ThreadMapMutex() );
	return !ThreadMap().empty();
};

Thread::~Thread()
{
}

Thread::Thread( pthread_t tid ) : thread( tid )
{
}

void*
Thread::launch( launch_info* oli )
{
	if ( !oli )
		_uds_abort;
	launch_info& li = *oli;
	
	{
		MutexLock lock( ThreadMapMutex() );
		MapInsert< internal_error >( ThreadMap(), pthread_self(), li._this );
	}
	
	void* ret;
	li._this->AddReference();
	
	pthread_cleanup_push( reinterpret_cast< vf_ptr >(
		static_cast< void ( * )( launch_info* ) >( &Delete ) ),
						  oli );
	pthread_cleanup_push( reinterpret_cast< vf_ptr >( &deactivate ),
						  li._this );
	try
	{
		ret = li.start();
	}
	catch ( Exit_t& e )
	{
		ret = reinterpret_cast< void* >( e.ReturnValue() );
	}
	pthread_cleanup_pop( 1 );
	pthread_cleanup_pop( 1 );
	
	return ret;
}

void
Thread::deactivate( Thread* _this )
{
	while ( !_this->chs.empty() )
		_this->PopCleanupHandler( true );
	
	ThreadMapMutex().Lock();
	ThreadMap().erase( pthread_self() );
	ThreadMapMutex().UnLock();
	
	_this->RemoveReference();
}


// Thread::launch_info implementation

Thread::launch_info::launch_info( const Routine& s, Thread* t ) : start( s ),
	_this( t )
{
}

}
